/* Author: Ruud Harmsen, 14-?? November 2023

   See https://rudhar.com/musica/hwedwortehun/ia14.htm for explanations.
   @? Not yet written
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "../../rhmusicfuncs/rhmusicfuncs.h"
#include "../../rhmusicfuncs/rhmidifuncs.h"
#include "../../rhmusicfuncs/rhmusicadhoc.h"
#include "../../rhmusicfuncs/rhmidibendbuf.h"

static char *outdir = "";
static int  verbose = 0;
static char *ProgPath = NULL;

static void Usage (void);
static int  AnalyseCmdlineArgs (int *p_argc, char** argv);
static char *MakeOutpath (char *MidiFilename);
static int  CopyToOutfile (char *MidiFilename, char *Outpath);
static void DoOneFile (char *MidiFilename);
static void DoTheFile (char *MidiFilename, FILE *fpi, FILE *fpo);
static void DoTheNotes (
   FILE *fpi, FILE *fpo, char *MidiFilename, long *TrackStarts,
   int NumTracks, int TicksPerQuarter);

/***********************************************************************/
int main (int argc, char **argv)
{
   ProgPath = argv[0];

   argv += AnalyseCmdlineArgs(&argc, argv);
   for ( ; argc >= 1; argc--, argv++)
   {
      DoOneFile(argv[0]);
   }
   return 0;
}

/***********************************************************************/
static void Usage (void)
{
   fprintf(stderr, "\nUsage: %s [-v[v[v]]] [-o outdir] [midi-file ...]\n\n", ProgPath);
   fprintf(stderr, "-v Verbose. Two of three v's causes more verbosity.\n");
   fprintf(stderr, "-o Specifies an output directory, that must pre-exist.\n");
   fprintf(stderr, "   The modified midi files, with \"adhoc\" in the name,\n");
   fprintf(stderr, "   before the name file extension (.mid or .midi), will\n");
   fprintf(stderr, "   be placed there. Default: same as the input file.\n");
}

/***********************************************************************/
static int AnalyseCmdlineArgs (int *p_argc, char** argv)
{
   int used_up = 0;

   /* Skip program name */
   for (used_up++, (*p_argc)--, argv++; *p_argc > 0; (*p_argc)--, argv++)
   {
      if (strncmp(argv[0], "-v", 2) == 0)
      {
         verbose = 1;
         if (argv[0][2] == 'v') verbose++;
         if (argv[0][3] == 'v') verbose++;
         used_up++;
      }
      else if (strcmp(argv[0], "-o") == 0 && *p_argc > 1)
      {
         outdir = argv[1];
         (*p_argc)--, argv++;
         used_up += 2;
      }
      else if (argv[0][0] == '-')
      {
         fprintf(stderr, "%s: Invalid argument: %s\n", ProgPath, argv[0]);
         Usage();
      }
      else if (argv[0][0] != '-')
         break;
   }
   return used_up;
}

/***********************************************************************/
static char *MakeOutpath (char *MidiFilename)
{
   size_t lenout = strlen(outdir);
   size_t to_allocate;
   int addslash = 0;
   char *Outpath;
   char *Marker = "-adhoc";
   char *WhereDot, *LastSlash;

   if (strstr(MidiFilename, Marker))
   /* Looks like the result of an earlier run was selected
      with a shell wildcard. Silently ignore, as converting
      this file again probably makes no sense. */
      return NULL;

   if (lenout > 0 && outdir[lenout - 1] != '/')
      addslash = 1;
   to_allocate =
      lenout + addslash + strlen(MidiFilename) + strlen(Marker) + 1;
   Outpath = calloc(1, to_allocate);
   if (!Outpath)
   {
      fprintf(stderr, "Source file %s, line %d\n", __FILE__, __LINE__);
      fprintf(stderr, "%s: Cannot calloc %ld bytes for outpath\n",
         ProgPath, (long)to_allocate);

      return Outpath;
   }
   if (MidiFilename[0] == '/')
      sprintf(Outpath, "%s", MidiFilename);
   else
      sprintf(Outpath, "%s%s%s", outdir, addslash ? "/" : "", MidiFilename);

   WhereDot  = strrchr(Outpath, '.');
   LastSlash = strrchr(Outpath, '/');
   if (!WhereDot || WhereDot < LastSlash)
   {
      fprintf(stderr,
         "%s: Midi filename %s does not have a filename extension\n",
         ProgPath, MidiFilename);
      fprintf(stderr, "Should be %s.midi or %s.mid etc.\n",
         MidiFilename, MidiFilename);
      Usage();

      return NULL;
   }
   memmove(WhereDot + strlen(Marker), WhereDot, strlen(Marker));
   memcpy (WhereDot, Marker, strlen(Marker));

   return Outpath;
}

/***********************************************************************/
static int CopyToOutfile (char *MidiFilename, char *Outpath)
{
   FILE *fpi, *fpo;
   int c;

   if (!(fpi = fopen(MidiFilename, "r")))
   {
      fprintf(stderr, "%s: Cannot open input file %s\n",
         ProgPath, MidiFilename);
      Usage();
      return 1;
   }
   if (!(fpo = fopen(Outpath, "w")))
   {
      fprintf(stderr, "%s: Cannot open output file %s\n",
         ProgPath, Outpath);
      Usage();
      return 2;
   }

   while ((c = getc(fpi)) != EOF)
      putc(c, fpo);

   fclose(fpi);
   fclose(fpo);

   return 0;
}

/***********************************************************************/
static void DoOneFile (char *MidiFilename)
{
   char *Outpath = MakeOutpath(MidiFilename);
   FILE *fpi = NULL, *fpo = NULL;

   if (!Outpath)
      return;

   if (CopyToOutfile(MidiFilename, Outpath))
   {
      free(Outpath);
      return;
   }

   if (!(fpi = fopen(MidiFilename, "r")))
      fprintf(stderr, "%s: Cannot open %s for reading\n",
         ProgPath, MidiFilename);
   else if (!(fpo = fopen(Outpath, "r+")))
      fprintf(stderr, "%s: Cannot open %s for reading and writing\n",
         ProgPath, Outpath);
   else
      DoTheFile(MidiFilename, fpi, fpo);

   if (fpi) fclose(fpi);
   if (fpo) fclose(fpo);
   free(Outpath);
}

/***********************************************************************/
static void DoTheFile (char *MidiFilename, FILE *fpi, FILE *fpo)
{
   int NumTracks;
   long WhereNumTracks;
   long *TrackStarts;
   int TicksPerQuarter;
   int err = 0;

   if (verbose >= 1)
      fprintf(stdout, "\n%s: Opened midi file %s for reading\n",
         ProgPath, MidiFilename);

   TrackStarts = ReadMidiHeadTrackstarts(
      fpi, &NumTracks, &WhereNumTracks, &TicksPerQuarter, &err);
   if (!TrackStarts)
   {
      if (err != 5)
         fprintf(stderr,
            "%s: No valid or usable header/tracks found in midifile %s\n",
            ProgPath, MidiFilename);
      switch (err)
      {
         case 1: fprintf(stderr,
            "%s: File %s has fewer bytes than needed for a midi header\n",
               ProgPath, MidiFilename);
            break;
         case 2: fprintf(stderr,
            "%s: File %s does not start with the MThd midi header marker\n",
               ProgPath, MidiFilename);
            break;
         case 3: fprintf(stderr,
            "%s: File %s: only Midi formats 0 and 1 supported\n",
               ProgPath, MidiFilename);
            break;
         case 4: fprintf(stderr,
            "%s: File %s: only ticks per quarter-note supported, not time codes\n",
               ProgPath, MidiFilename);
            break;
         case 5: fprintf(stderr,
            "%s: File %s: can't calloc space for start positions of %d midi tracks\n",
               ProgPath, MidiFilename, NumTracks);
            break;
         case 6: fprintf(stderr,
            "%s: File %s: didn't find the start of a Track where it was expected\n",
               ProgPath, MidiFilename); fprintf(stderr,
            "%s: File position (after MTrk and length) was %ld bytes (hex %lx)\n",
               ProgPath, ftell(fpi), ftell(fpi));
            break;
      }
      return;
   }

   if (verbose > 2)
   {
      fprintf(stdout, "%s: File %s: %2d tracks at pos %ld\n",
         ProgPath, MidiFilename, NumTracks, WhereNumTracks);
      for (int i = 0; i < NumTracks; i++)
         fprintf(stdout, "%s: File %s: Track %2d starts at %6ld (hex %5lx)\n",
            ProgPath, MidiFilename, i, TrackStarts[i], TrackStarts[i]);
   }
   AdhocInitliseBendTrack(fpo, WhereNumTracks, NumTracks);
   DoTheNotes(fpi, fpo, MidiFilename, TrackStarts, NumTracks, TicksPerQuarter);
   AdhocFinaliseBendTrack(fpo);

   free(TrackStarts);
}

/***********************************************************************/
static void DiagPerTrack (
   int NumTracks, FILE *fpi, char *ProgPath, char *MidiFilename,
   long *TrackStarts, struct rhmidi_eventdata **EvtdatTracks,
   int TicksPerQuarter, int verbose)
{
   /* For diagnostic purposes only. */
   int track;

   for (track = 0; track < NumTracks; track++)
   {
      fprintf(stdout, "\nTrack %2d\n", track);
      while (ReadMidiEventErrHandling(fpi, ProgPath, MidiFilename,
             EvtdatTracks[track], TicksPerQuarter, verbose) >= 0)
      {
         ;
      }
   }
   fprintf(stdout, "End of per-track listing\n\n");
   /* Reset for doing what is really the purpose of the program. */
   for (track = 0; track < NumTracks; track++)
   {
      memset(EvtdatTracks[track], '\0', sizeof (struct rhmidi_eventdata));
      EvtdatTracks[track]->TrackFilepos = TrackStarts[track];
      EvtdatTracks[track]->TrackNumber  = track;
   }
}

/***********************************************************************/
static void DoTheNotes (
   FILE *fpi, FILE *fpo, char *MidiFilename, long *TrackStarts,
   int NumTracks, int TicksPerQuarter)
{
   int track;
   struct rhmidi_eventdata **EvtdatTracks;

   /* Allocate data space for all the tracks. */
   if (!(EvtdatTracks = calloc(NumTracks, sizeof (struct rhmidi_eventdata *))))
   {
      fprintf(stderr,
         "%s: In file %s: cannot malloc data buffer. Src %s line %d\n",
            ProgPath, MidiFilename, __FILE__, __LINE__);
      return;
   }
   for (track = 0; track < NumTracks; track++)
   {
      if (!(EvtdatTracks[track] = calloc(1, sizeof (struct rhmidi_eventdata))))
      {
         fprintf(stderr,
            "%s: In file %s: cannot malloc data buffer. Src %s line %d\n",
               ProgPath, MidiFilename, __FILE__, __LINE__);
         return;
      }
      EvtdatTracks[track]->TrackFilepos = TrackStarts[track];
      EvtdatTracks[track]->TrackNumber  = track;
   }

   if (verbose > 2)
      DiagPerTrack(NumTracks, fpi, ProgPath, MidiFilename,
         TrackStarts, EvtdatTracks, TicksPerQuarter, verbose);

   AdhocInitBuffer();
   EvtdatTracks[0]->EndOfTrack = 0;
   while (1)
   {
      AdhocRefillBendBuf(fpi, ProgPath, MidiFilename,
         EvtdatTracks, NumTracks, TicksPerQuarter, verbose);
      AdhocFillChordArray(verbose);
      AdhocRunAdhocAlgorithm(verbose);
      if (AdhocWriteBendEvents(fpo) == EOF)
         break;
   }

   /* Free directly and indirectly allocated memory blocks. */
   for (track = 0; track < NumTracks; track++)
   {
      free(EvtdatTracks[track]->EventData);
      free(EvtdatTracks[track]);
   }
   free(EvtdatTracks);
}
